"""
pycycling is a cross-platform Python package for interacting with various Bluetooth cycling peripherals.

The package provides a number of classes which wrap around `Bleak <https://github.com/hbldh/bleak>`_ client objects.

.. _obtaining_device_address:

Obtaining the address of your device
====================================

This documentation includes a number of code snippets demonstrating package usage. Each of these has a
hardcoded device address ID which you will need to replace with the address of your device. The following script can
be used to obtain this address:

.. literalinclude:: ../examples/discover_devices_example.py
"""
